@@ -2,6 +2,7 @@ from __future__ import annotations
 from collections import deque, Counter, OrderedDict
 from contextlib import contextmanager
 import json, time, random, logging
+from sqlite3.dbapi2 import ProgrammingError
 from json import encoder
 import sqlite3
 from types import FunctionType
@@ -38,8 +39,8 @@ class SQLiteDB:
         try:
             with self.new_connection() as con:
                 return con.execute(query, params or [])
-        except sqlite3.OperationalError as e:
-            e.args = (*e.args, query, str(params))
+        except sqlite3.Error as e:
+            e.args = (*e.args, query, params)
             raise e
 
     def execute(self, query: str, params: Params = None) -> List[Any]:
@@ -229,7 +230,7 @@ class IndexedSQLiteTable(SQLiteTable):
                        f' ({v}) vs ({kwargs[c]})')
                 raise Exception(msg)
             kwargs[c] = v
-        if not self.update_row(idx, **kwargs):
+        if not self.update_row(*idx, **kwargs):
             self.insert_row(**kwargs)
         return
 
@@ -250,7 +251,7 @@ class KeyValueSQLiteTable(SQLiteTable):
         self.db.execute(f'''
         CREATE TABLE IF NOT EXISTS {self.table_name}(
             key text NOT NULL PRIMARY KEY,
-            value text NOT NULL,
+            value text,
             lock_token double NOT NULL,
             locked_until double NOT NULL
         )
@@ -260,13 +261,48 @@ class KeyValueSQLiteTable(SQLiteTable):
         return
 
     @contextmanager
-    def exclusive_access(self, *keys: str, delta=0.02, max_duration=0.5):
+    def get_lock(self, *keys: str, max_duration: float = 0.5,
+                 request_every: float = 0.02, timeout: Optional[float] = 10):
+        '''
+        # Wait at most {timeout} seconds to get exclusive access,
+        # requesting every {request_every} seconds
+        with table.get_lock('some_key'):
+            current = table['some_key']
+            ...
+            # No one else can set table['some_key'] = something_else
+            # during next max_duration seconds
+            ...
+            table['some_key'] = some_value
+        '''
         assert keys, 'You must specify keys to be locked explicitely'
         token = 1 + random.random()
         for key in keys:
-            self._lock(key, token, delta, max_duration)
+            self._get_lock(key, token=token, max_duration=max_duration,
+                           request_every=request_every, timeout=timeout)
         try:
-            yield
+            yield True
+        finally:
+            for key in keys:
+                self._unlock(key, token, max_duration)
+        return
+
+    @contextmanager
+    def ask_lock(self, *keys: str, max_duration: float = 0.5):
+        '''
+        with table.ask_lock('some_key') as gained_access:
+            if gained_access:
+                # The resource is mine
+            else:
+                # The resource is locked by other
+        '''
+        assert keys, 'You must specify keys to be locked explicitely'
+        token = 1 + random.random()
+        gained_access = all([
+            self._ask_lock(key, token=token, max_duration=max_duration)
+            for key in keys
+        ])
+        try:
+            yield gained_access
         finally:
             for key in keys:
                 self._unlock(key, token, max_duration)
@@ -276,22 +312,34 @@ class KeyValueSQLiteTable(SQLiteTable):
         lock_keys = ['lock_token', 'locked_until']
         return super().unique_dict(lock_keys, 'key=?', [key])
 
-    def _lock(self, key: str, token: float, delta=0.02, max_duration=1.5):
-        while True:
-            d = self._current_lock(key)
-            if (not d or d['lock_token'] == 0 or
-                    time.time() > d['locked_until']):
-                # Request access, race until next loop
-                super().indexed_by(['key']).set_row(
-                    key,
-                    lock_token=token,
-                    locked_until=time.time() + max_duration,
-                )
-            elif d['lock_token'] == token:
-                break  # Access gained
-            time.sleep(delta)
+    def _get_lock(self, key: str, token: float, max_duration: float,
+                  request_every: float, timeout: float = None):
+        deadline = float('inf') if timeout is None else time.time() + timeout
+        while not self._ask_lock(key, token, max_duration):
+            if time.time() > deadline:
+                raise Exception('TIMEOUT')
+            else:
+                time.sleep(request_every)
         return
 
+    def _ask_lock(self, key: str, token: float, max_duration: float):
+        d = self._current_lock(key)
+        may_request = (not d or d['lock_token'] == 0 or
+                       time.time() > d['locked_until'] or
+                       d['lock_token'] == token)
+        if may_request:
+            # Request access, race until next read
+            super().indexed_by(['key']).set_row(
+                key,
+                lock_token=token,
+                locked_until=time.time() + max_duration,
+            )
+            # Read and check
+            d = self._current_lock(key)
+            if d and d['lock_token'] == token:
+                return True  # Access gained
+        return False
+
     def _unlock(self, key: str, token: float, max_duration: float):
         d = self._current_lock(key)
         if not d:
@@ -313,21 +361,21 @@ class KeyValueSQLiteTable(SQLiteTable):
         d = super().unique_dict(['value'], 'key=?', [key])
         if not d:
             raise KeyError(key)
-        return json.loads(d['value'])
+        return json.loads(d['value'] or 'null')
 
     def __setitem__(self, key: str, value: Any):
         super().indexed_by(['key']).set_row(key, value=json.dumps(value))
         return
 
     def values(self):
-        return [json.loads(s) for s in super().column('value')]
+        return [json.loads(s or 'null') for s in super().column('value')]
 
     def keys(self) -> List[str]:
         return [k for k in super().column('key')]
 
     def items(self) -> List[Tuple[str, Any]]:
         items = super().rows(['key', 'value'])
-        return [(k, json.loads(v)) for k, v in items]
+        return [(k, json.loads(v or 'null')) for k, v in items]
 
     def get(self, key: str, default=None):
         try:
@@ -355,4 +403,9 @@ def test():
     print(len(ob))
     ob.set_row('key1', obj='VALUE1')
     print(ob.get_row('key1'))
-    print(len(ob))
\ No newline at end of file
+    print(len(ob))
+    table = db.get_table('test1').as_key_value()
+    with table.get_lock('last_time', 'last_progress'):
+        print('I HAVE ACCESS!')
+        with table.ask_lock('last_time') as access:
+            assert not access
